/*=========================================================================
  
  Authors: Daniel Aguilera, Thierry Carrard, Claire Guilbaud
  
  Copyright (c) CEA/DIF (Comissariat a l'Energie Atomique, Direction Ile de France) 
                BP12 91680 Bruyeres Le Chatel 
                France 
  All rights reserved.
  
     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.
  
=========================================================================*/


#include "vtkVRPNCameraManipulator.h"
#include "vtkVRPNCommand.h"

#include <vtkRenderWindow.h>
#include <vtkRenderWindowInteractor.h>
#include <vtkCallbackCommand.h>
#include <vtkCamera.h>
#include <vtkTransform.h>
#include <vtkObjectFactory.h>

#include <vrpn_Tracker.h>
#include <vrpn_Analog.h>
#include <quat.h>

vtkCxxRevisionMacro(vtkVRPNCameraManipulator,"1.0");
vtkStandardNewMacro(vtkVRPNCameraManipulator);

vtkVRPNCameraManipulator::vtkVRPNCameraManipulator()
  : _scale(1.0),
    Camera(0),
    _camTransform(0),
    TrackerSensor(0),
    _resetPosition(false)
{
  _camTransform = vtkTransform::New();
  for(int i=0;i<3;i++) _startPosition[i] = 0.0;
  for(int i=0;i<6;i++) Bounds[i] = (i%2==0) ? -1.0 : 1.0 ;
  for(int i=0;i<3;i++) AnalogTranslateChannels[i] = i;
  for(int i=0;i<3;i++) AnalogRotateChannels[i] = i;
}

vtkVRPNCameraManipulator::~vtkVRPNCameraManipulator()
{
}

void vtkVRPNCameraManipulator::SetCamera(vtkCamera* cam)
{
  if(cam!=Camera)
  {
    Camera = cam;
    this->ResetCamera();
  }
}

void vtkVRPNCameraManipulator::ResetCamera()
{
  if( Camera!=0 )
  {
    Camera->GetPosition( _position );
    Camera->GetFocalPoint( _focal );
    Camera->GetViewUp( _viewUp );
    
    _resetPosition = true;
  }
}

void vtkVRPNCameraManipulator::ProcessEvent(vtkObject* caller, unsigned long event, void* calldata)
{
  vrpn_TRACKERCB* trk;
  vrpn_ANALOGCB* anlg;

  switch(event)
    {
      case vtkVRPNCommand::VRPN_TRACKER :
	trk = (vrpn_TRACKERCB *) calldata;
	if(trk->sensor==TrackerSensor && Camera!=0)
	{
	  q_type quat;
	  double orient[16];
	  double pos[3];
	  double focal[4];
	  double viewUp[4];
	  double tmp[4];

	  // emulate tracker reset, record position when resetPosition flag is set
	  if( _resetPosition )
	  {
	    for(int i=0;i<3;i++) _startPosition[i] = trk->pos[i];
	    q_copy(_startOrient, trk->quat);
	    q_invert(_startOrient, _startOrient);
	    _resetPosition = false;
	  }

	  // compute relative position from start position
	  for(int i=0;i<3;i++) pos[i] = trk->pos[i] - _startPosition[i];

	  // compute relative orientation from start orientation
	  q_mult( quat, trk->quat, _startOrient );
	  q_invert(quat,quat);

	  // build rotation matrix
	  q_to_ogl_matrix(orient, quat);

	  // compute new camera position
	  for(int i=0;i<3;i++) pos[i] = _position[i]+pos[i]*_scale;

	  // rotate focal point around original camera position and offset it to new position
	  for(int i=0;i<3;i++) tmp[i] = _focal[i] - _position[i];
	  tmp[3]=1.0;
	  vtkMatrix4x4::MultiplyPoint(orient,tmp,focal);
	  for(int i=0;i<3;i++) focal[i] = (focal[i]/focal[3]) + pos[i];
	  
	  // rotate view Up
	  for(int i=0;i<3;i++) tmp[i] = _viewUp[i];
	  tmp[3]=1.0;
	  vtkMatrix4x4::MultiplyPoint(orient,tmp,viewUp);
	  for(int i=0;i<3;i++) viewUp[i] /= viewUp[3];

	  Camera->SetPosition( pos );
	  Camera->SetFocalPoint( focal );
	  Camera->SetViewUp( viewUp );
	  Camera->Modified();
	}
	break;
      case vtkVRPNCommand::VRPN_ANALOG :
	{
	  anlg = (vrpn_ANALOGCB*) calldata;
	  double tx,ty,tz,rx,ry,rz;
	  for(int i=0;i<anlg->num_channel;i++)
	  {
		cerr<<anlg->channel[i]<<' ';
	  }
	  cerr<<endl; cerr.flush();

#define GetAnalogValue(n,i) \
(Analog##n##Channels[i]<0||Analog##n##Channels[i]>=anlg->num_channel)?0.0:anlg->channel[Analog##n##Channels[i]];

	  tx = GetAnalogValue(Translate,0);
	  ty = GetAnalogValue(Translate,1);
	  tz = GetAnalogValue(Translate,2);
	  rx = GetAnalogValue(Rotate,0);
	  ry = GetAnalogValue(Rotate,1);
	  rz = GetAnalogValue(Rotate,2);

	  _camTransform->Identity();
	  _camTransform->Translate(tx,ty,tz);
	  _camTransform->RotateX(rx);
	  _camTransform->RotateY(ry);
	  _camTransform->RotateZ(rz);
	  Camera->ApplyTransform(_camTransform);
	}
	break;
    case vtkCommand::CharEvent :
      cout<<"CharEvent: "<<caller->GetClassName()<<endl;
      break;
    case vtkCommand::KeyPressEvent :
      cout<<"KeyPressEvent: "<<caller->GetClassName()<<endl;
      break;
    }
}
