/*=========================================================================
  
  Authors: Daniel Aguilera, Thierry Carrard, Claire Guilbaud
  
  Copyright (c) CEA/DIF (Comissariat a l'Energie Atomique, Direction Ile de France) 
                BP12 91680 Bruyeres Le Chatel 
                France 
  All rights reserved.
  
     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.
  
=========================================================================*/


#include "vtkVRPNWandManipulator.h"
#include "vtkVRPNCommand.h"

#include <vtkConeSource.h>
#include <vtkPolyDataMapper.h>
#include <vtkActor.h>
#include <vtkTransform.h>
#include <vtkMatrix4x4.h>
#include <vtkRenderWindow.h>
#include <vtkProp3D.h>
#include <vtkObjectFactory.h>
#include <vtkRenderer.h>
#include <vtkRendererCollection.h>
#include <vtkRenderWindow.h>
#include <vtkObjectFactory.h>
#include <vrpn_Tracker.h>
#include <vrpn_Button.h>
#include <quat.h>

vtkCxxRevisionMacro(vtkVRPNWandManipulator,"1.0");
vtkStandardNewMacro(vtkVRPNWandManipulator);

vtkVRPNWandManipulator::vtkVRPNWandManipulator()
  : MoveThreshold(0),
    _scale(1.0),
    _state(FREE_MOVE),
    TrackerSensor(0),
    Button(0),
    _renwin(0),
    _prop(0)
{
  _objTransform = vtkTransform::New();
  _wandTransform = vtkTransform::New();
  vtkMatrix4x4::Identity( _wandMatrix );
  vtkMatrix4x4::Identity( _objMatrix );

  // Wand avatar
  vtkPolyDataMapper* coneMapper;
  _cone = vtkConeSource::New();
  _cone->SetResolution(20);
  coneMapper = vtkPolyDataMapper::New();
  coneMapper->SetInput(_cone->GetOutput());
  _coneActor = vtkActor::New();
  _coneActor->SetMapper(coneMapper);
  _coneActor->SetUserTransform(_wandTransform);
}

vtkVRPNWandManipulator::~vtkVRPNWandManipulator()
{
}

void vtkVRPNWandManipulator::SetProp3D(vtkProp3D* p)
{
  double bounds[6];
  _prop = p;
  _scale = -1;
  _prop->GetBounds (bounds);
  for(int i=0;i<3;i++)
  {
    double s = bounds[i*2+1] - bounds[i*2];
    if(s<0) s = -s;
    if( _scale==-1 || s>_scale ) _scale = s;
  }
  _cone->SetHeight(_scale*0.2);
  _cone->SetRadius(_scale*0.1);

  _objTransform->Identity();
  vtkMatrix4x4::Identity( _objMatrix );
  _prop->SetUserTransform(_objTransform);
}

void vtkVRPNWandManipulator::SetRenderWindow(vtkRenderWindow* rw)
{
  if(rw!=_renwin)
  {
    if(_renwin!=0)
    {
      _renwin->GetRenderers()->GetFirstRenderer()->RemoveProp( _coneActor );      
    }    
    _renwin = rw;
    if(_renwin!=0)
    {
      _renwin->GetRenderers()->GetFirstRenderer()->AddProp( _coneActor );
    }
  }
}

static double MoveQuantity(const double invM1[16], const double M2[16])
{
  double tmp[16];
  double id[16];
  double sum = 0.0;

  vtkMatrix4x4::Multiply4x4( M2, invM1, tmp );
  vtkMatrix4x4::Identity(id);
  for(int i=0;i<16;i++)
  {
    double e = tmp[i] - id[i];
    sum += e*e;
  }
  return sum;
}


void vtkVRPNWandManipulator::DragStart()
{
  if( _state == FREE_MOVE )
  {
    vtkMatrix4x4::Invert( _wandMatrix, _invStartWandMatrix );
    vtkMatrix4x4::DeepCopy( _invPauseWandMatrix, _invStartWandMatrix );
    _state = DRAG_START;
  }
}

void vtkVRPNWandManipulator::Move()
{
  _wandTransform->Identity();
  _wandTransform->Concatenate(_wandMatrix);
  if( _state == FREE_MOVE ) return;

  double move = MoveQuantity( _invPauseWandMatrix, _wandMatrix );
  if( _state == DRAG_START )
  {
    if(move>=MoveThreshold)
    {
      _state = DRAG_MOVE;
    }
  }
  else
  {
    // move /= elapsed_time;
    if(move<MoveThreshold)
    {
      _state = DRAG_PAUSE;
    }
  }

  if( _state == DRAG_MOVE )
  {
    vtkMatrix4x4::Invert( _wandMatrix, _invPauseWandMatrix );
    _objTransform->Identity();
    _objTransform->Scale(_scale,_scale,_scale);
    _objTransform->Concatenate( _wandMatrix );
    _objTransform->Concatenate( _invStartWandMatrix );
    _objTransform->Concatenate( _objMatrix );
  }
}

void vtkVRPNWandManipulator::DragEnd()
{
  if( _state==DRAG_MOVE || _state==DRAG_PAUSE )
  {
    double move[16];
    double final[16];
    vtkMatrix4x4::Multiply4x4( _wandMatrix, _invStartWandMatrix, move );
    vtkMatrix4x4::Multiply4x4( move, _objMatrix, final );
    vtkMatrix4x4::DeepCopy( _objMatrix, final );
    _state = FREE_MOVE;
  }
}

void vtkVRPNWandManipulator::ProcessEvent(vtkObject* caller, unsigned long event, void* calldata)
{
  vrpn_TRACKERCB* trackPosInfo;
  vrpn_BUTTONCB* buttonInfo;

  switch(event)
    {
      case vtkVRPNCommand::VRPN_TRACKER :
	trackPosInfo = (vrpn_TRACKERCB*) calldata;
	if(trackPosInfo->sensor==TrackerSensor)
	{
	  double translate[16];
	  vtkMatrix4x4::Identity(translate);
	  translate[12] = trackPosInfo->pos[0];
	  translate[13] = trackPosInfo->pos[1];
	  translate[14] = trackPosInfo->pos[2];

	  double orient[16];
	  q_to_ogl_matrix(orient, trackPosInfo->quat);
	  
	  vtkMatrix4x4::Multiply4x4( translate, orient, _wandMatrix );
	  this->Move();
	}
	break;

      case vtkVRPNCommand::VRPN_BUTTON:	
	buttonInfo = (vrpn_BUTTONCB*) calldata;
	if(buttonInfo->button==Button)
	{
	  if(buttonInfo->state) this->DragStart();
	  else this->DragEnd();
	}
    }
}
